/*
 * Copyright (c) 2021 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/logging/log.h>
#include <stdio.h>
#include <stdlib.h>
#include <psa/crypto.h>
#include <psa/crypto_extra.h>

#include "base64.h"
#include "base64_2.h"

#ifdef CONFIG_BUILD_WITH_TFM
#include <tfm_ns_interface.h>
#endif

#define APP_SUCCESS   (0)
#define APP_ERROR   (-1)
#define APP_SUCCESS_MESSAGE "Example finished successfully!"
#define APP_ERROR_MESSAGE "Example exited with error!"

#define PRINT_HEX(p_label, p_text, len)\
  ({\
    LOG_INF("---- %s (len: %u): ----", p_label, len);\
    LOG_HEXDUMP_INF(p_text, len, "Content:");\
    LOG_INF("---- %s end  ----", p_label);\
  })

LOG_MODULE_REGISTER(rsa, LOG_LEVEL_DBG);

/* ====================================================================== */
/*        Global variables/defines for the RSA example        */

#define NRF_CRYPTO_EXAMPLE_RSA_TEXT_SIZE (100)
//#define NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_SIZE (140)
#define NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_SIZE (280)
#define NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_BASE64_SIZE (200)
#define NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_SIZE (128)
#define NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_BASE64_SIZE (512)

/* Below text is used as plaintext for signing using RSA . */
static char m_plain_text[NRF_CRYPTO_EXAMPLE_RSA_TEXT_SIZE] = {
  "Example string to demonstrate basic usage of RSA."
};

static char m_pub_key[NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_SIZE];
ssize_t m_pub_key_length;
static char m_pub_key_base64[NRF_CRYPTO_EXAMPLE_RSA_PUBLIC_KEY_BASE64_SIZE];
ssize_t m_pub_key_base64_length;

static char m_signature[NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_SIZE];
static char m_signature_base64[NRF_CRYPTO_EXAMPLE_RSA_SIGNATURE_BASE64_SIZE];
ssize_t m_signature_base64_length;
static char m_hash[32];

static psa_key_handle_t keypair_handle;
static psa_key_handle_t pub_key_handle;
/* ====================================================================== */

#define RSA_KEY \
"MIIBCwIBADANBgkqhkiG9w0BAQEFAASB9jCB8wIBAAIxALvDGoG4fxARGmIOS2OR" \
"Hu5pFhMleB5oXoZwkth4Y3ZZEwxa3NHlDJFeKlh325Yz5wIDAQABAjAe0NKU37ba" \
"6DOwTDBLUxIEhvK8ibXBJ9ezMj5zudTlZkfb75hOJkbukFiJdFz9rGECGQD3fKrQ" \
"iXmXlYOmuVv6buEElGXvkUT9I8MCGQDCOIIeA3ZoWfNOmtVXESJj9qEL9iSv4Q0C" \
"GQCBlWIvHLxw1VWAl6+EWkgxyK3zGWJBFVcCGG9gFE290pL4sivrYlfxfXUxoxzG" \
"5DbvqQIZAJbLfxZSY3G0PT8yYi/WECNjm8qBTokPQQ=="
#define RSA_KEY_PUBLIC \
"MEwwDQYJKoZIhvcNAQEBBQADOwAwOAIxALvDGoG4fxARGmIOS2ORHu5pFhMleB5o" \
"XoZwkth4Y3ZZEwxa3NHlDJFeKlh325Yz5wIDAQAB"
#define STRING_TO_SIGN "nRF9160_cc1310nrf91603526561072349682.00nrf9160_352656107234968"
#define EXPCTED_OUTPUT "Sn6GGeyhi9OGPjjtIceQdFrSZ5ndLIFxSMuucvXWIoeLT1vBu+OiCjlEOddY20Gg"

int crypto_init(void)
{
  psa_status_t status;

  /* Initialize PSA Crypto */
  status = psa_crypto_init();
  if (status != PSA_SUCCESS)
    return APP_ERROR;

  return APP_SUCCESS;
}

int crypto_finish(void)
{
  psa_status_t status;

  /* Destroy the key handle */
  status = psa_destroy_key(keypair_handle);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_destroy_key failed! (Error: %d)", status);
    return APP_ERROR;
  }

  status = psa_destroy_key(pub_key_handle);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_destroy_key failed! (Error: %d)", status);
    return APP_ERROR;
  }

  return APP_SUCCESS;
}

int generate_rsa_keypair(void)
{
  psa_status_t status;
  size_t olen;

  LOG_INF("Generating random RSA keypair...");

  /* Configure the key attributes */
  psa_key_attributes_t key_attributes = PSA_KEY_ATTRIBUTES_INIT;

  /* Configure the key attributes */
  psa_set_key_usage_flags(&key_attributes, PSA_KEY_USAGE_SIGN_HASH);
  psa_set_key_lifetime(&key_attributes, PSA_KEY_LIFETIME_VOLATILE);
  psa_set_key_algorithm(&key_attributes, PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_256));
  psa_set_key_type(&key_attributes, PSA_KEY_TYPE_RSA_KEY_PAIR);
  psa_set_key_bits(&key_attributes, 1024);

  /* Generate a random keypair. The keypair is not exposed to the application,
   * we can use it to signing/verification the key handle.
   */
  status = psa_generate_key(&key_attributes, &keypair_handle);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_generate_key failed! (Error: %d)", status);
    return APP_ERROR;
  }

  /* Export the public key */
  status = psa_export_public_key(keypair_handle, m_pub_key, sizeof(m_pub_key), &olen);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_export_public_key failed! (Error: %d)", status);
    return APP_ERROR;
  }

  /* After the key handle is acquired the attributes are not needed */
  psa_reset_key_attributes(&key_attributes);

  LOG_INF("RSA generated successfully!");

  return APP_SUCCESS;
}

int import_rsa_pub_key(void)
{
  /* Configure the key attributes */
  psa_key_attributes_t key_attributes = PSA_KEY_ATTRIBUTES_INIT;
  psa_status_t status;

  /* Configure the key attributes */
  psa_set_key_usage_flags(&key_attributes, PSA_KEY_USAGE_VERIFY_HASH);
  psa_set_key_lifetime(&key_attributes, PSA_KEY_LIFETIME_VOLATILE);
  psa_set_key_algorithm(&key_attributes, PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_256));
  psa_set_key_type(&key_attributes, PSA_KEY_TYPE_RSA_PUBLIC_KEY);
  psa_set_key_bits(&key_attributes, 1024);
  /*
  PRINT_HEX("m_pub_key_before", m_pub_key, sizeof(m_pub_key));

  m_pub_key_base64_length = encode_base64(m_pub_key, sizeof(m_pub_key), m_pub_key_base64);
  m_pub_key_base64[m_pub_key_base64_length] = 0;
  LOG_INF("m_pub_key_base64 = %s", m_pub_key_base64);
  LOG_INF("m_pub_key_base64_length = %d", m_pub_key_base64_length);
  //PRINT_HEX("m_pub_key_base64", m_pub_key_base64, m_pub_key_base64);

  m_pub_key_length = decode_base64_with_length(m_pub_key_base64, m_pub_key_base64_length, m_pub_key);
  LOG_INF("m_pub_key_length = %d", m_pub_key_length);
  PRINT_HEX("m_pub_key_after", m_pub_key, sizeof(m_pub_key));
  */
  m_pub_key_length = decode_base64_with_length(RSA_KEY, sizeof(RSA_KEY), m_pub_key);
  LOG_INF("m_pub_key_length = %d", m_pub_key_length);
  PRINT_HEX("m_pub_key", m_pub_key, sizeof(m_pub_key));

  status = psa_import_key(&key_attributes, m_pub_key, sizeof(m_pub_key), &pub_key_handle);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_import_key failed! (Error: %d)", status);
    return APP_ERROR;
  }

  /* After the key handle is acquired the attributes are not needed */
  psa_reset_key_attributes(&key_attributes);

  return APP_SUCCESS;
}

int sign_message_rsa(void)
{
  uint32_t olen;
  psa_status_t status;

  LOG_INF("Signing a message using RSA...");

  /* Compute the SHA256 hash */
  status = psa_hash_compute(
    PSA_ALG_SHA_256, m_plain_text, sizeof(m_plain_text), m_hash, sizeof(m_hash), &olen);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_hash_compute failed! (Error: %d)", status);
    return APP_ERROR;
  }

  /* Sign the hash using RSA */
  status = psa_sign_hash(keypair_handle,
             PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_256),
             m_hash,
             sizeof(m_hash),
             m_signature,
             sizeof(m_signature),
             &olen);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_sign_hash failed! (Error: %d)", status);
    return APP_ERROR;
  }

  LOG_INF("Signing was successful!");
  PRINT_HEX("Plaintext", m_plain_text, sizeof(m_plain_text));
  PRINT_HEX("SHA256 hash", m_hash, sizeof(m_hash));
  PRINT_HEX("Signature", m_signature, sizeof(m_signature));

  return APP_SUCCESS;
}

int sign_message_rsa2(void)
{
  uint32_t olen;
  psa_status_t status;

  LOG_INF("Signing a message using RSA...");

  /* Compute the SHA256 hash */
  status = psa_hash_compute(
    PSA_ALG_SHA_256, m_plain_text, sizeof(m_plain_text), m_hash, sizeof(m_hash), &olen);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_hash_compute failed! (Error: %d)", status);
    return APP_ERROR;
  }

  /* Sign the hash using RSA */
  status = psa_sign_message(keypair_handle,
             //PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_256),
             PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_1),
             STRING_TO_SIGN,
             sizeof(STRING_TO_SIGN),
             m_signature,
             sizeof(m_signature),
             &olen);
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_sign_message failed! (Error: %d)", status);
    return APP_ERROR;
  }

  LOG_INF("Signing was successful!");
  PRINT_HEX("Plaintext", m_plain_text, sizeof(m_plain_text));
  PRINT_HEX("SHA256 hash", m_hash, sizeof(m_hash));
  PRINT_HEX("Signature", m_signature, sizeof(m_signature));

  m_signature_base64_length = encode_base64(m_signature, sizeof(m_signature), m_signature_base64);
  m_signature_base64[m_signature_base64_length] = 0;
  LOG_INF("m_signature_base64 = %s", m_signature_base64);
  LOG_INF("m_signature_base64_length = %d", m_signature_base64_length);

  return APP_SUCCESS;
}

int verify_message_rsa(void)
{
  psa_status_t status;

  LOG_INF("Verifying RSA signature...");

  /* Verify the hash */
  status = psa_verify_hash(pub_key_handle,
         PSA_ALG_RSA_PKCS1V15_SIGN(PSA_ALG_SHA_256),
         m_hash,
         sizeof(m_hash),
         m_signature,
         sizeof(m_signature));
  if (status != PSA_SUCCESS) {
    LOG_INF("psa_verify_hash failed! (Error: %d)", status);
    return APP_ERROR;
  }

  LOG_INF("Signature verification was successful!");

  return APP_SUCCESS;
}

int rsa_test(void)
{
  int status;

  LOG_INF("Starting the RSA example...");

  status = crypto_init();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  status = generate_rsa_keypair();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  status = import_rsa_pub_key();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  //status = sign_message_rsa();
  status = sign_message_rsa2();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  status = verify_message_rsa();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  status = crypto_finish();
  if (status != APP_SUCCESS) {
    LOG_INF(APP_ERROR_MESSAGE);
    return APP_ERROR;
  }

  LOG_INF(APP_SUCCESS_MESSAGE);

  return APP_SUCCESS;
}

int main(void) {
  rsa_test();

  return APP_SUCCESS;
}